#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneController.py
 @brief Control the flight of the drone.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 BSD 3clause "New" or "Revised" License

"""
import sys
import math

# Import RTM module
import RTC
import OpenRTM_aist

import target_gps
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# constants
BUTTON_01_MASK = 0x0001
BUTTON_02_MASK = 0x0002
BUTTON_03_MASK = 0x0004
BUTTON_04_MASK = 0x0008
BUTTON_05_MASK = 0x0010
BUTTON_06_MASK = 0x0020
BUTTON_07_MASK = 0x0040
BUTTON_08_MASK = 0x0080
BUTTON_09_MASK = 0x0100
BUTTON_10_MASK = 0x0200
BUTTON_11_MASK = 0x0400
BUTTON_12_MASK = 0x0800
BUTTON_13_MASK = 0x1000
BUTTON_14_MASK = 0x2000

MOVE_SPEED_MAX = 10.0
MOVE_SPEED_MIN = 1.0

MOVE_UPDATE_COEF_ABSOLUTE_NS = 1.0
MOVE_UPDATE_COEF_ABSOLUTE_EW = 0.65
MOVE_UPDATE_COEF_RELATIVE = 0.55

ALLOWABLE_ERROR_POS_CMP = 5.0

ALLOWABLE_ERROR_HEIGHT = 0.0

MODE_ABSOLUTE = 1

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
dronecontroller_spec = ["implementation_id", "DroneController", 
         "type_name",         "DroneController", 
         "description",       "Control the flight of the drone.", 
         "version",           "1.2.1", 
         "vendor",            "EAC", 
         "category",          "Drone", 
         "activity_type",     "STATIC", 
         "max_instance",      "1", 
         "language",          "Python", 
         "lang_type",         "SCRIPT",
         "conf.default.AltCoefficient", "1",
         "conf.default.LatLonCoefficient", "10",
         "conf.default.LatDiff10m", "0.000137",
         "conf.default.LonDiff10m", "0.000108",
         "conf.default.MoveSpeed", "3.0",
         "conf.default.MinAlt", "2.0",
         "conf.default.MaxAlt", "15.0",
         "conf.default.TurnSpeed", "45",
         "conf.default.NoseLatDiff10m", "9.01323",
         "conf.default.NoseLonDiff10m", "8.9831566",
         "conf.default.MoveTurnSpeed", "30",
         "conf.__widget__.AltCoefficient", "spin",
         "conf.__widget__.LatLonCoefficient", "spin",
         "conf.__widget__.LatDiff10m", "spin",
         "conf.__widget__.LonDiff10m", "spin",
         "conf.__widget__.MoveSpeed", "spin",
         "conf.__widget__.MinAlt", "spin",
         "conf.__widget__.MaxAlt", "spin",
         "conf.__widget__.TurnSpeed", "spin",
         "conf.__widget__.NoseLatDiff10m", "spin",
         "conf.__widget__.NoseLonDiff10m", "spin",
         "conf.__widget__.MoveTurnSpeed", "spin",
         "conf.__constraints__.AltCoefficient", "1<=x<=10",
         "conf.__constraints__.LatLonCoefficient", "2<=x<=10",
         "conf.__constraints__.LatDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.LonDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.MoveSpeed", "1.0<=x<=10.0",
         "conf.__constraints__.MinAlt", "1.0<=x<=100.0",
         "conf.__constraints__.MaxAlt", "10.0<=x<=100.0",
         "conf.__constraints__.TurnSpeed", "0<=x<=360",
         "conf.__constraints__.NoseLatDiff10m", "0.000000<=x<=10.000000",
         "conf.__constraints__.NoseLonDiff10m", "0.0000000<=x<=10.0000000",
         "conf.__constraints__.MoveTurnSpeed", "0<=x<=360",
         ""]
# </rtc-template>

##
# @class DroneController
# @brief Control the flight of the drone.
# 
# Refer to the "機能仕様書_DroneController" for details.
# 
# Refer to the "機能仕様書_DroneController" for details.
# 
# Refer to the "機能仕様書_DroneController" for details.
# 
# Refer to the "機能仕様書_DroneController" for details.
# 
# 
class DroneController(OpenRTM_aist.DataFlowComponentBase):
    
    ##
    # @brief constructor
    # @param manager Maneger Object
    # 
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_HomeGPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                        0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ホーム（基地局）GPSデータ
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._HomeGPSInIn = OpenRTM_aist.InPort("HomeGPSIn", self._d_HomeGPSIn)
        self._d_DroneGPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                        0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ドローンGPSデータ
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._DroneGPSInIn = OpenRTM_aist.InPort("DroneGPSIn", self._d_DroneGPSIn)
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0,0),0)
        """
        PS4コントローラのボタン入力状態
         - Type: RTC::TimedULong
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._ButtonInIn = OpenRTM_aist.InPort("ButtonIn", self._d_ButtonIn)
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
        """
        PS4コントローラのアナログ入力状態
         - Type: RTC::TimedDoubleSeq
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._AnalogInIn = OpenRTM_aist.InPort("AnalogIn", self._d_AnalogIn)
        self._d_inMag = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        磁気コンパス
         - Type: RTC::TimedFloat
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inMagIn = OpenRTM_aist.InPort("inMag", self._d_inMag)
        self._d_TargetGPSOut = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                        0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        移動ターゲットGPSデータ
         - Type: RTC::GPSTime
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._TargetGPSOutOut = OpenRTM_aist.OutPort("TargetGPSOut", self._d_TargetGPSOut)
        self._d_outTargetSpeed = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        ターゲット速度データ
         - Type: RTC::TimedFloat
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTargetSpeedOut = OpenRTM_aist.OutPort("outTargetSpeed", self._d_outTargetSpeed)
        self._d_outTargetYaw = RTC.TimedShortSeq(RTC.Time(0,0),[])
        """
        旋回方向及び移動量
         - Type: RTC::TimedShortSeq
         - Semantics: data[0]:旋回方向(1:右　-1:左　0:旋回無し)
                      data[1]:ターゲット絶対角(単位:deg)
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTargetYawOut = OpenRTM_aist.OutPort("outTargetYaw", self._d_outTargetYaw)
        self._d_outMoveMode = RTC.TimedString(RTC.Time(0,0),0)
        """
        移動モード(Absolute or Relative)
         - Type: RTC::TimedString
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outMoveModeOut = OpenRTM_aist.OutPort("outMoveMode", self._d_outMoveMode)


        


        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        移動ターゲット高度算出係数
         - Name: AltCoefficient AltCoefficient
         - DefaultValue: 1
         - Range: 1<=x<=10
        """
        self._AltCoefficient = [1]
        """
        移動ターゲット経緯度算出係数
         - Name: LatLonCoefficient LatLonCoefficient
         - DefaultValue: 10
         - Range: 2<=x<=10
        """
        self._LatLonCoefficient = [10]
        """
        10mあたりの緯度差分
         - Name: LatDiff10m LatDiff10m
         - DefaultValue: 0.000137
         - Range: 0.000000<x<0.100000
        """
        self._LatDiff10m = [0.000137]
        """
        10mあたりの経度差分
         - Name: LonDiff10m LonDiff10m
         - DefaultValue: 0.000108
         - Range: 0.000000<x<0.100000
        """
        self._LonDiff10m = [0.000108]
        """
        ドローン移動速度
         - Name: Speed MoveSpeed
         - DefaultValue: 3.0
         - Range: 1.0<=x<=10.0
        """
        self._MoveSpeed = [3.0]
        """
        ドローン最小高度
         - Name: MinAlt MinAlt
         - DefaultValue: 2.0
         - Range: 1.0<=x<=100.0
        """
        self._MinAlt = [2.0]
        """
        ドローン最大高度
        ※ドローン最大高度(MaxAlt)の設定値は、最小高度(MinAlt)の設定値以上の値とする
        。
         - Name: MaxAlt MaxAlt
         - DefaultValue: 15.0
         - Range: 10.0<=x<=100.0
        """
        self._MaxAlt = [15.0]
        """
        ドローン旋回速度
         - Name: TurnSpeed TurnSpeed
         - DefaultValue: 45
         - Range: 0<=x<=360
        """
        self._TurnSpeed = [45]
        """
        10mあたりの機首緯度差分
         - Name: NoseLatDiff10m NoseLatDiff10m
         - DefaultValue: 9.01323
         - Range: 0.000000<=x<=10.000000
        """
        self._NoseLatDiff10m = [9.01323]
        """
        10mあたりの機首経度差分
         - Name: NoseLonDiff10m NoseLonDiff10m
         - DefaultValue: 8.9831566
         - Range: 0.0000000<=x<=10.0000000
        """
        self._NoseLonDiff10m = [8.9831566]
        """
        ドローン移動旋回速度
         - Name: MoveTurnSpeed MoveTurnSpeed
         - DefaultValue: 30
         - Range: 0<=x<=360
        """
        self._MoveTurnSpeed = [30]
        
        # </rtc-template>

        self._target_gps = target_gps.TargetGPS()

        self._home_gps = False
        self._drone_gps = False
        self._button = False
        self._analog = False
        self._mag = False

        self._SpeedParam = self._MoveSpeed[0]
        
        self._last_front_back = 0
        self._last_right_left = 0
        self._last_up_down = 0
        self._last_speedparam = 0
        self._last_move_mode = "Absolute"

        # connection check for data port ButtonInIn
        func = lambda: self._cb_connect_status_change("ButtonIn", True)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("ButtonIn", False)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnDisconnected(ob)

        # connection check for data port AnalogInIn
        func = lambda: self._cb_connect_status_change("AnalogIn", True)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("AnalogIn", False)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnDisconnected(ob)
         
    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry() 
    # 
    # @return RTC::ReturnCode_t
    # 
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("AltCoefficient", self._AltCoefficient, "1")
        self.bindParameter("LatLonCoefficient", self._LatLonCoefficient, "10")
        self.bindParameter("LatDiff10m", self._LatDiff10m, "0.000137")
        self.bindParameter("LonDiff10m", self._LonDiff10m, "0.000108")
        self.bindParameter("MoveSpeed", self._MoveSpeed, "3.0")
        self.bindParameter("MinAlt", self._MinAlt, "2.0")
        self.bindParameter("MaxAlt", self._MaxAlt, "15.0")
        self.bindParameter("TurnSpeed", self._TurnSpeed, "45")
        self.bindParameter("NoseLatDiff10m", self._NoseLatDiff10m, "9.01323")
        self.bindParameter("NoseLonDiff10m", self._NoseLonDiff10m, "8.9831566")
        self.bindParameter("MoveTurnSpeed", self._MoveTurnSpeed, "30")
        
        # Set InPort buffers
        self.addInPort("HomeGPSIn",self._HomeGPSInIn)
        self.addInPort("DroneGPSIn",self._DroneGPSInIn)
        self.addInPort("ButtonIn",self._ButtonInIn)
        self.addInPort("AnalogIn",self._AnalogInIn)
        self.addInPort("inMag",self._inMagIn)
        
        # Set OutPort buffers
        self.addOutPort("TargetGPSOut",self._TargetGPSOutOut)
        self.addOutPort("outTargetSpeed",self._outTargetSpeedOut)
        self.addOutPort("outTargetYaw",self._outTargetYawOut)
        self.addOutPort("outMoveMode",self._outMoveModeOut)
        
        # Set service provider to Ports
        
        # Set service consumers to Ports
        
        # Set CORBA Service Ports
        
        return RTC.RTC_OK
    
    #    ##
    #    # 
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    # 
    #    # @return RTC::ReturnCode_t
    #
    #    # 
    #def onFinalize(self):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    # 
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # 変数初期化
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        # 
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):
        init_movemode = MODE_ABSOLUTE
        self._target_gps.set_controller_mode(init_movemode)
        self._SpeedParam = self._MoveSpeed[0]

        self._last_move_mode = "Absolute"
    
        return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The deactivated action (Active state exit action)
    #    # former rtc_active_exit()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onDeactivated(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # PS4キー入力、ドローンステータス取得、およびコントロールデータ送信の周期処理。
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
        self._read_button()
        self._read_analog()
        self._read_home_gps()
        self._read_drone_gps()
        self._read_mag()

        self._set_target_speed()
        self._set_move_mode()
        self._set_target_yaw()
        self._set_target_gps()
        self._set_outport()
  
        return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onError(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    #def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    def _read_button(self):
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)

        if self._button:
            self._d_ButtonIn = self._ButtonInIn.read()
            logger.debug("Button = 0b" + format(self._d_ButtonIn.data, '016b'))
        else:
            logger.error("ERROR: ButtonIn is NOT connected")

    def _read_analog(self):
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])

        if self._analog:
            self._d_AnalogIn = self._AnalogInIn.read()
            for i in range(0, len(self._d_AnalogIn.data), 2):
                d1 = d2 = None
                try:
                    d1 = self._d_AnalogIn.data[i]
                    d2 = self._d_AnalogIn.data[i + 1]
                except IndexError:
                    pass

                if d2 is None:
                    logger.debug("Analog[%d]:  % f", i, d1)
                else:
                    logger.debug("Analog[%d,%d]: % f % f", i, i + 1, d1, d2)
        else:
            logger.error("ERROR: AnalogIn is NOT connected")

    def _read_home_gps(self):
        self._home_gps = False

        if self._HomeGPSInIn.isNew():
            self._d_HomeGPSIn = self._HomeGPSInIn.read()
            self._home_gps = True
            logger.debug("HomeGPS:   % f % f % f",
                         self._d_HomeGPSIn.latitude,
                         self._d_HomeGPSIn.longitude,
                         self._d_HomeGPSIn.altitude)

    def _read_drone_gps(self):
        self._drone_gps = False

        if self._DroneGPSInIn.isNew():
            self._d_DroneGPSIn = self._DroneGPSInIn.read()
            self._drone_gps = True
            logger.debug("DroneGPS:  % f % f % f",
                         self._d_DroneGPSIn.latitude,
                         self._d_DroneGPSIn.longitude,
                         self._d_DroneGPSIn.altitude)

    def _read_mag(self):
        self._mag = False

        if self._inMagIn.isNew():
            self._d_inMag = self._inMagIn.read()
            self._mag = True
            logger.debug("Mag: %d",
                         self._d_inMag.data)

    def _config_target_gps(self):
        config = {"AltCoefficient": self._AltCoefficient[0],
                  "LatLonCoefficient": self._LatLonCoefficient[0],
                  "LatDiff10m": self._LatDiff10m[0],
                  "LonDiff10m": self._LonDiff10m[0],
                  "MoveSpeed": self._MoveSpeed[0],
                  "MinAlt": self._MinAlt[0],
                  "MaxAlt": self._MaxAlt[0],
                  "TurnSpeed": self._TurnSpeed[0],
                  "NoseLatDiff10m": self._NoseLatDiff10m[0],
                  "NoseLonDiff10m": self._NoseLonDiff10m[0],
                  "MoveTurnSpeed": self._MoveTurnSpeed[0]
                  }

        self._target_gps.config_param(config)

    def _set_target_gps(self):
        if not self._home_gps & self._drone_gps:
            logger.error("ERROR: GPS data is NOT available")
            return

        self._config_target_gps()

        target = None
        target_update = True
        takeoff_landing = self._takeoff_landing()
        if takeoff_landing:
            target_update = True
            target = self._target_gps. \
                get_takeoff_landing_data(self._d_HomeGPSIn, takeoff_landing)
        else:
            mv = self._moving()
            move_continue = self._move_continue(mv.front_back, mv.right_left, mv.up_down)
            if move_continue:
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist = dist_cal.distance_math(self._d_TargetGPSOut.longitude,
                                              self._d_DroneGPSIn.longitude,
                                              self._d_TargetGPSOut.latitude,
                                              self._d_DroneGPSIn.latitude,
                                              0,
                                              0)

                if self._d_outMoveMode.data == "Absolute":
                    if mv.front_back:
                        update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_NS
                    else:
                        update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_EW
                elif self._d_outMoveMode.data == "Relative":
                    update_coefficient = MOVE_UPDATE_COEF_RELATIVE

                if self._last_speedparam == self._SpeedParam:
                    if dist >= (self._SpeedParam * self._LatLonCoefficient[0] * update_coefficient):
                        target_update = False

            if target_update:
                target = self._target_gps.get_move_target_data(self._d_DroneGPSIn,
                                                               mv.front_back,
                                                               mv.right_left,
                                                               mv.up_down,
                                                               self._SpeedParam,
                                                               self._d_inMag.data)

            self._last_front_back = mv.front_back
            self._last_right_left = mv.right_left
            self._last_up_down = mv.up_down

        if target_update:
            self._d_TargetGPSOut.latitude = target.latitude
            self._d_TargetGPSOut.longitude = target.longitude
            self._d_TargetGPSOut.altitude = target.altitude

    def _set_target_speed(self):
        self._move_speed()
        self._d_outTargetSpeed.data = self._SpeedParam

    def _set_target_yaw(self):
        if self._mag:
            target_deg = 0  
            target_dir = int(self._turn_dir())
            mv1 = self._front_back()
            mv2 = self._right_left()
            mv3 = self._up_down()

            move_continue = self._move_continue(mv1, mv2, mv3)
            if move_continue == False:
                self._relative_mag = int(self._d_inMag.data)

            if target_dir:
                if mv1 or mv2 or mv3:
                    turn_speed = self._MoveTurnSpeed[0]
                else:
                    turn_speed = self._TurnSpeed[0]

                target_deg = int(self._d_inMag.data) + (turn_speed * target_dir)
            else:
                if mv1 or mv2:
                    if self._d_outMoveMode.data == "Absolute":
                        if self._last_front_back or self._last_right_left:
                            target_deg = int(self._diff_target_angle())
                            if target_deg > 0:
                                target_dir = 1
                            else:
                                target_dir = -1
                            target_deg += int(self._d_inMag.data)

                    elif self._d_outMoveMode.data == "Relative":
                        target_dir = 1
                        target_deg = self._relative_mag


            if target_deg > 360:
                target_deg = target_deg % 360
            elif target_deg < 0:
                target_deg = 360 - (abs(target_deg) % 360)

            self._d_outTargetYaw.data = [target_dir,target_deg]

        else:
            logger.error("ERROR: MAG data is NOT available")

    def _set_move_mode(self):
        mode = self._move_mode()

        if mode == 1:
            self._d_outMoveMode.data = "Absolute"
        else:
            self._d_outMoveMode.data = "Relative"

        self._target_gps.set_controller_mode(mode)

    def _set_outport(self):
        if self._home_gps & self._drone_gps:
            self._TargetGPSOutOut.write()
            logger.debug("TargetGPS: % f % f % f",
                     self._d_TargetGPSOut.latitude,
                     self._d_TargetGPSOut.longitude,
                     self._d_TargetGPSOut.altitude)

        self._outTargetSpeedOut.write()
        logger.debug("TargetSpeed: %f",
                     self._d_outTargetSpeed.data)
        self._last_speedparam = self._SpeedParam

        if self._mag:
            self._outTargetYawOut.write()
            logger.debug("TargetYaw: (dir)%d, (deg)%d",
                        self._d_outTargetYaw.data[0],
                        self._d_outTargetYaw.data[1])

        self._outMoveModeOut.write()
        logger.debug("MoveMode: %s",
                     self._d_outMoveMode.data)
        self._last_move_mode = self._d_outMoveMode.data

    def _diff_target_angle(self):
        angle_vec_x = math.cos(math.radians(self._d_inMag.data))
        angle_vec_y = math.sin(math.radians(self._d_inMag.data))

        target_vec_x = self._d_TargetGPSOut.latitude - self._d_DroneGPSIn.latitude
        target_vec_y = self._d_TargetGPSOut.longitude - self._d_DroneGPSIn.longitude

        inner_product = angle_vec_x * target_vec_x + angle_vec_y * target_vec_y
        cross_product = angle_vec_x * target_vec_y - angle_vec_y * target_vec_x

        angle_rad = math.atan2(cross_product, inner_product)
        angle_deg = math.degrees(angle_rad)

        return angle_deg

    def _check_drone_on_home(self):
        ret = False

        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(self._d_HomeGPSIn.longitude,
                                      self._d_DroneGPSIn.longitude,
                                      self._d_HomeGPSIn.latitude,
                                      self._d_DroneGPSIn.latitude,
                                      0,
                                      0)

        # TODO: 最適な誤差をシステムテストで割り出す
        if (dist <= ALLOWABLE_ERROR_POS_CMP and
                self._d_DroneGPSIn.altitude < (self._MinAlt[0] - ALLOWABLE_ERROR_HEIGHT)):

            ret = True

        return ret

    def _takeoff_landing(self):
        operation = 0
        home = self._check_drone_on_home()
        if home:
            # check takeoff request
            if self._d_ButtonIn.data & BUTTON_06_MASK:
                operation = 1
            elif self._d_ButtonIn.data & BUTTON_08_MASK:
                logger.warning("WARNING: Landing operation is "
                               "NOT available "
                               "in the current drone position")
        else:
            # check landing request
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                operation = -1
            elif self._d_ButtonIn.data & BUTTON_06_MASK:
                logger.warning("WARNING: Take-off operation is "
                               "NOT available in the current drone position")

        return operation

    def _move_mode(self):
        home = self._check_drone_on_home()
        mode = self._target_gps._mode
        
        if home:
            if self._d_ButtonIn.data & BUTTON_10_MASK:
                logger.warning("WARNING: MoveMode operation is "
                               "NOT available in the current drone position")                
        else:
            if self._d_ButtonIn.data & BUTTON_10_MASK:
                if mode == 1:
                    mode = 2
                else:
                    mode = 1

        return mode

    def _moving(self):
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        home = self._check_drone_on_home()

        if home:
            if mv1 or mv2 or mv3:
                mv1 = mv2 = mv3 = 0
                logger.warning("WARNING: Move operations are "
                               "NOT available in the current drone position")

        return self.MovingOperation(mv1, mv2, mv3)

    def _front_back(self):
        key = 0
        try:
            key = self._d_AnalogIn.data[0]
        except IndexError:
            pass

        operation = self._front_back_check(key)
        operation *= -1

        return operation

    def _right_left(self):
        operation = 0
        both_push = BUTTON_03_MASK | BUTTON_01_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [right/left]")
        elif self._d_ButtonIn.data & BUTTON_03_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_01_MASK:
            operation = -1

        return operation

    def _up_down(self):
        operation = 0
        both_push = BUTTON_04_MASK | BUTTON_02_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [up/down]")
        elif self._d_ButtonIn.data & BUTTON_04_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_02_MASK:
            operation = -1

        return operation

    def _move_continue(self, front_back, right_left, up_down):
        if self._last_move_mode == self._d_outMoveMode.data:
            if self._turn_dir():
                move_continue = False
            else:
                if (front_back != self._last_front_back) or (right_left != self._last_right_left) or (up_down != self._last_up_down):
                    move_continue = False
                elif (front_back == 0) and (right_left == 0) and (up_down == 0):
                    move_continue = False
                else:
                    move_continue = True
        else:
            move_continue = False

        return move_continue

    def _turn_dir(self):
        home = self._check_drone_on_home()
        key = 0

        try:
            key = self._d_AnalogIn.data[1]
        except IndexError:
            pass

        if home:
            if key:
                key = 0
                logger.warning("WARNING: Turning operation is "
                               "NOT available in the current drone position")

        else:
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                key = 0
            else:
                if self._d_outMoveMode.data == "Absolute":
                    mv1 = self._front_back()
                    mv2 = self._right_left()
                    mv3 = self._up_down()
                    if mv1 or mv2 or mv3:
                        key = 0

        operation = self._turn_dir_check(key)

        return operation

    def _move_speed(self):
        home = self._check_drone_on_home()
        both_push = BUTTON_05_MASK | BUTTON_07_MASK
        
        if home:
            if (self._d_ButtonIn.data & BUTTON_05_MASK) or (self._d_ButtonIn.data & BUTTON_07_MASK) or (self._d_ButtonIn.data & both_push):
                logger.warning("WARNING: Speed operation is "
                               "NOT available in the current drone position") 
        elif (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [MoveSpeed up/down]")
        elif self._d_ButtonIn.data & BUTTON_05_MASK:
            self._SpeedParam += 1.0

        elif self._d_ButtonIn.data & BUTTON_07_MASK:
            self._SpeedParam -= 1.0

        self._SpeedParam = self._move_speed_check(self._SpeedParam)

    def _front_back_check(self, status):
        if status > 1:
            status = 1
        elif status < -1:
            status = -1
            
        return status

    def _turn_dir_check(self, dir):
        if dir > 1:
            dir = 1
        elif dir < -1:
            dir = -1
            
        return dir

    def _move_speed_check(self, speed):
        if speed > MOVE_SPEED_MAX:
            speed = MOVE_SPEED_MAX
        elif speed < MOVE_SPEED_MIN:
            speed = MOVE_SPEED_MIN
            
        return speed
    

    def _cb_connect_status_change(self, port, connect):
        if port == "ButtonIn":
            self._button = connect
        elif port == "AnalogIn":
            self._analog = connect
        else:
            logger.warning("Invalid port name: %s", port)

        if not connect:
            logger.warning("ERROR: Data port was disconnected: %s", port)

    class MovingOperation(object):

        def __init__(self, front_back, right_left, up_down):
            self.front_back = front_back
            self.right_left = right_left
            self.up_down = up_down


class ConnectionCallback(OpenRTM_aist.PortCallBack.ConnectionCallback):

    def __init__(self, func):
        self._plugin_func = func

    def __call__(self, profile):
        self._plugin_func()


def DroneControllerInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=dronecontroller_spec)
    manager.registerFactory(profile,
                            DroneController,
                            OpenRTM_aist.Delete)

def MyModuleInit(manager):
    DroneControllerInit(manager)

    # Create a component
    comp = manager.createComponent("DroneController")

def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()

