#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneViewer.py
 @brief Display a variety of info of the drone
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 BSD 3-clause "New" or "Revised" License

"""
import sys
import time
import datetime

# Import RTM module
import RTC
import OpenRTM_aist
import create_sensor_panel
import math
import drone_distance_cal as dr_cal

import Tkinter

sys.path.append(".")

drone_view = create_sensor_panel.CreateSensorPanel()

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
droneviewer_spec = ["implementation_id", "DroneViewer",
                    "type_name",         "DroneViewer",
                    "description",       "Display a variety of info of the drone",
                    "version",           "1.2.0",
                    "vendor",            "EAC",
                    "category",          "Viewer",
                    "activity_type",     "STATIC",
                    "max_instance",      "1",
                    "language",          "Python",
                    "lang_type",         "SCRIPT",
                    ""]
# </rtc-template>

##
# @class DroneViewer
# @brief Display a variety of info of the drone
#
# Overview: Refer to the "機能仕様書_DroneViewer.pdf" for details.
#
# In / Out: Refer to the "機能仕様書_DroneViewer.pdf" for details.
#
# Algorythm: Refer to the "機能仕様書_DroneViewer.pdf" for details.
#
#


class DroneViewer(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inGPSHome = RTC.GPSData(
            RTC.Time(0, 0), RTC.GPSTime(0, 0), 0.0, 0.0, 0.0,
            0.0, 0.0, 0.0, 0.0, 0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        基地局の位置情報を取得する。
         - Type: RTC::GPSData
         - Number: 3 (latitude/longitude/altitude)
         - Semantics: [latitude]
                      緯度(DEG形式)
                      [longitude]
                      経度(DEG形式)
                      [altitude]
                      高度(単位:m)
         - Unit: -
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inGPSHomeIn = OpenRTM_aist.InPort("inGPSHome", self._d_inGPSHome)
        self._d_inGPSDrone = RTC.GPSData(
            RTC.Time(0, 0), RTC.GPSTime(0, 0), 0.0, 0.0, 0.0, 0.0,
            0.0, 0.0, 0.0, 0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        Drone本体の位置情報を取得する。
         - Type: RTC::GPSData
         - Number: 3
         - Semantics: [latitude]
                      緯度(DEG形式)
                      [longitude]
                      経度(DEG形式)
                      [altitude]
                      高度(単位:m)
         - Unit: -
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inGPSDroneIn = OpenRTM_aist.InPort(
            "inGPSDrone", self._d_inGPSDrone)
        self._d_inBattV = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        バッテリー電圧値
         - Type: RTC::TimedFloat
         - Number: 1
         - Semantics: -
         - Unit: mA
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inBattVIn = OpenRTM_aist.InPort("inBattV", self._d_inBattV)
        self._d_inBattmA = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        バッテリー電流値
         - Number: 1
         - Unit: mA
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inBattmAIn = OpenRTM_aist.InPort("inBattmA", self._d_inBattmA)
        self._d_inFltTime = RTC.Time(RTC.Time(0, 0), 0)
        """
        Flight時間情報を取得する。
         - Number: 1
         - Unit: min:sec
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inFltTimeIn = OpenRTM_aist.InPort("inFltTime", self._d_inFltTime)
        self._d_inGyro = RTC.TimedOrientation3D(
            RTC.Time(0, 0),
            RTC.Orientation3D(0.0, 0.0, 0.0))
        """
        Gyro情報を取得する。
         - Number: 3
         - Semantics: roll
                      pitch
                      yaw
         - Unit: rad
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inGyroIn = OpenRTM_aist.InPort("inGyro", self._d_inGyro)
        self._d_inAccel = RTC.TimedFloatSeq(RTC.Time(0, 0), [])
        """
        加速度情報を取得する。
         - Number: 3
         - Semantics: [data]
                      0:x
                      1:y
                      2:z
         - Unit: m/s
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inAccelIn = OpenRTM_aist.InPort("inAccel", self._d_inAccel)
        self._d_inMag = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        磁気コンパス情報を取得する
         - Number: 1
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inMagIn = OpenRTM_aist.InPort("inMag", self._d_inMag)
        self._d_inTemp = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        気温情報を取得する
         - Number: 1
         - Unit: C
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inTempIn = OpenRTM_aist.InPort("inTemp", self._d_inTemp)
        self._d_inPres = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        気圧情報を取得する。
         - Number: 1
         - Unit: hPa
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inPresIn = OpenRTM_aist.InPort("inPres", self._d_inPres)
        self._d_inFltMode = RTC.TimedString(RTC.Time(0, 0), 0)
        """
        現在のフライトモードを取得する。
         - Number: 1
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inFltModeIn = OpenRTM_aist.InPort("inFltMode", self._d_inFltMode)
        self._d_inState = RTC.TimedString(RTC.Time(0, 0), 0)
        """
        ドローンI/Oの内部モードを取得する
         - Number: 1
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inStateIn = OpenRTM_aist.InPort("inState", self._d_inState)
        self._d_inSysErr = RTC.TimedString(RTC.Time(0, 0), 0)
        """
        ドローンI/Oの内部エラーメッセージを取得する
         - Number: 1
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inSysErrIn = OpenRTM_aist.InPort("inSysErr", self._d_inSysErr)
        self._d_inTargetSpeed = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        ターゲット速度を取得する。
         - Number: 1
         - Unit: m/s
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inTargetSpeedIn = OpenRTM_aist.InPort(
            "inTargetSpeed", self._d_inTargetSpeed)
        self._d_inMoveMode = RTC.TimedString(RTC.Time(0, 0), 0)
        """
        現在の移動モードを取得する。
         - Number: 1
         - Frequency: 10
         - Operation Cycle: 10
        """
        self._inMoveModeIn = OpenRTM_aist.InPort(
            "inMoveMode", self._d_inMoveMode)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">

        # </rtc-template>

    ##
    # Initialization of RTC
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    # @pre nothing special
    # @post nothing special
    #
    def onInitialize(self):
        # Bind variables and configuration variable

        # Set InPort buffers
        self.addInPort("inGPSHome", self._inGPSHomeIn)
        self.addInPort("inGPSDrone", self._inGPSDroneIn)
        self.addInPort("inBattV", self._inBattVIn)
        self.addInPort("inBattmA", self._inBattmAIn)
        self.addInPort("inFltTime", self._inFltTimeIn)
        self.addInPort("inGyro", self._inGyroIn)
        self.addInPort("inAccel", self._inAccelIn)
        self.addInPort("inMag", self._inMagIn)
        self.addInPort("inTemp", self._inTempIn)
        self.addInPort("inPres", self._inPresIn)
        self.addInPort("inFltMode", self._inFltModeIn)
        self.addInPort("inState", self._inStateIn)
        self.addInPort("inSysErr", self._inSysErrIn)
        self.addInPort("inTargetSpeed", self._inTargetSpeedIn)
        self.addInPort("inMoveMode", self._inMoveModeIn)

        # Set OutPort buffers

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        self.GPSHome_readin = False
        self.GPSDrone_readin = False
        self.dis_bk = 0.0

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    #
    #    # @return RTC::ReturnCode_t
    #
    #    #
    # def onFinalize(self):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # Activate the RTC
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):

        self.GPSHome_readin = False
        self.GPSDrone_readin = False
        self.dis_bk = 0.0
        drone_view.sensor_sts_init()

        return RTC.RTC_OK

        ##
        # Deactivate the RTC
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        # @pre nothing special
        # @post nothing special
        #
    def onDeactivated(self, ec_id):

        return RTC.RTC_OK

        ##
        # Periodic task of RTC. This task send parameter to 2DCDP RTC.
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        # @pre nothing special
        # @post nothing special
        #
    def onExecute(self, ec_id):

        self.read_GPS_home()
        self.read_GPS_drone()
        self.read_battery()
        self.read_flighttime()
        self.read_gyro()
        self.read_accel()
        self.read_Mag()
        self.read_temperature()
        self.read_pressure()
        self.read_State()
        self.read_SysErr()
        self.read_fltmode()
        self.read_targetspd()
        self.read_movemode()
        self.calc_dist()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onError(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    def read_GPS_home(self):
        if self._inGPSHomeIn.isNew():
            data = self._inGPSHomeIn.read()

            self.hm_lat = data.latitude
            self.hm_lat = round(self.hm_lat, 6)

            self.hm_lon = data.longitude
            self.hm_lon = round(self.hm_lon, 6)

            self.hm_alt = data.altitude
            self.hm_alt = round(self.hm_alt, 1)

            drone_view.lan_var.set(str(self.hm_lat))
            drone_view.lon_var.set(str(self.hm_lon))
            drone_view.alt_var.set(str(self.hm_alt))
            self.GPSHome_readin = True

            print "GPS Home (lat,lon,alt):" \
                + str(self.hm_lat) + "," + str(self.hm_lon) + ","\
                + str(self.hm_alt)

    def read_GPS_drone(self):
        if self._inGPSDroneIn.isNew():
            data = self._inGPSDroneIn.read()

            self.dr_lat = data.latitude
            self.dr_lat = round(self.dr_lat, 6)

            self.dr_lon = data.longitude
            self.dr_lon = round(self.dr_lon, 6)

            self.dr_alt = data.altitude
            self.dr_alt = round(self.dr_alt, 1)

            drone_view.drlan_var.set(str(self.dr_lat))
            drone_view.drlon_var.set(str(self.dr_lon))
            drone_view.dralt_var.set(str(self.dr_alt))
            self.GPSDrone_readin = True

            print "GPS drone (lat,lon,alt):"\
                + str(self.dr_lat) + "," + str(self.dr_lon) + ","\
                + str(self.dr_alt)

    def calc_dist(self):
        if self.GPSHome_readin and self.GPSDrone_readin:
            d = dr_cal.DroneMathmetics()
            d_y = d.distance_math(self.hm_lon, self.dr_lon, self.hm_lat,
                                  self.dr_lat, 0, self.dr_alt)
            distance = round(d_y, 1)
            drone_view.distance_var.set(str(distance))

            if distance == self.dis_bk:
                pass
            else:
                print "home-drone distance:" + str(distance) + "[m]"
                self.dis_bk = distance

    def read_battery(self):
        batt_V = 0.0
        if self._inBattVIn.isNew():
            battV = self._inBattVIn.read()
            batt_V = round(battV.data, 1)
            drone_view.batt_V_var.set(str(batt_V))

        batt_mA = 0.0
        if self._inBattmAIn.isNew():
            battmA = self._inBattmAIn.read()
            batt_mA = round(battmA.data, 1)
            drone_view.batt_mA_var.set(str(batt_mA))

        print "Batt:" + str(batt_V) + "[V]" + str(batt_mA) + "[mA]"

    def read_flighttime(self):
        if self._inFltTimeIn.isNew():
            tm = self._inFltTimeIn.read()
            tm_min = tm.sec // 60
            tm_sec = tm.sec % 60
            tm_str = "{0:d}:{1:02d}".format(tm_min, tm_sec)

            drone_view.flight_time.set(tm_str)

            print "flight time: " + tm_str

    def read_gyro(self):
        if self._inGyroIn.isNew():
            Gyro = self._inGyroIn.read()

            roll = Gyro.data.r
            roll = round(roll, 6)

            pitch = Gyro.data.p
            pitch = round(pitch, 6)

            yaw = Gyro.data.y
            yaw = round(yaw, 6)

            drone_view.roll_var.set(str(roll))
            drone_view.pitch_var.set(str(pitch))
            drone_view.yaw_var.set(str(yaw))

            print "Gyro(roll,pitch,yaw):"\
                + str(roll) + "," + str(pitch) + "," + str(yaw)

    def read_accel(self):
        if self._inAccelIn.isNew():
            Accel = self._inAccelIn.read()
            x = Accel.data[0]
            x = round(x, 2)

            y = Accel.data[1]
            y = round(y, 2)

            z = Accel.data[2]
            z = round(z, 2)

            drone_view.x_var.set(str(x))
            drone_view.y_var.set(str(y))
            drone_view.z_var.set(str(z))

            print "Accel(x,y,z):"\
                + str(x) + "," + str(y) + "," + str(z) + "[m/s]"

    def read_Mag(self):
        if self._inMagIn.isNew():
            Mag = int(self._inMagIn.read().data)
            drone_view.mag_var.set(str(Mag))

            print "Mag:" + str(Mag)

    def read_temperature(self):
        if self._inTempIn.isNew():
            tempdata = self._inTempIn.read().data
            temp = round(tempdata, 1)
            drone_view.temp_var.set(str(temp))
            print "Temperature:" + str(temp) + "[C]"

    def read_State(self):
        if self._inStateIn.isNew():
            dronestate = self._inStateIn.read().data
            drone_view.dronestate_var.set(dronestate)
            print "<info>State:" + dronestate

    def read_SysErr(self):
        if self._inSysErrIn.isNew():
            dronesyserr = self._inSysErrIn.read().data
            #drone_view.text_sys_err.delete('1.0', Tkinter.END)
            drone_view.text_sys_err.insert(Tkinter.INSERT, dronesyserr)
            print "SysErr:" + dronesyserr

    def read_pressure(self):
        if self._inPresIn.isNew():
            pressdata = self._inPresIn.read().data
            hpress = round(pressdata, 1)
            drone_view.press_var.set(str(hpress))

            print "Pressure:" + str(hpress) + "[hPa]"

    def read_fltmode(self):
        if self._inFltModeIn.isNew():
            fltmode = self._inFltModeIn.read().data
            drone_view.fltmd_var.set(fltmode)
            if fltmode in ["GUIDED", "RTL"]:
                drone_view.label_fltmd_sts.configure(bg="white")
            elif fltmode == "LOITER":
                drone_view.label_fltmd_sts.configure(bg="green")
            elif fltmode == "STABILIZE":
                drone_view.label_fltmd_sts.configure(bg="yellow")
            elif fltmode == "AUTO":
                drone_view.label_fltmd_sts.configure(bg="red")
            else:
                drone_view.label_fltmd_sts.configure(bg="white")

            print "Flt mode:" + fltmode

    def read_targetspd(self):
        if self._inTargetSpeedIn.isNew():
            targetspd = self._inTargetSpeedIn.read().data
            drone_view.maxspeed_var.set(str(targetspd))

            print "Target Speed:" + str(targetspd) + "[m/s]"

    def read_movemode(self):
        if self._inMoveModeIn.isNew():
            movemode = self._inMoveModeIn.read().data
            drone_view.movemode_var.set(movemode)

            print "MoveMode:" + movemode


def DroneViewerInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=droneviewer_spec)
    manager.registerFactory(profile,
                            DroneViewer,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    DroneViewerInit(manager)

    # Create a component
    comp = manager.createComponent("DroneViewer")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager(True)

    drone_view.mainloop()

if __name__ == "__main__":
    main()
