// -*- Java -*-
/*!
 * @file  MiniSurveyorIOImpl.java
 * @brief MiniSurveyor flight instruction. Get and output the sensor info.
 * @date  $Date$
 *
 * @author East Japan Accounting Center Co, Ltd
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;

import RTC.GPSData;
import RTC.Orientation3D;
import RTC.TimedFloat;
import RTC.Time;
import RTC.TimedOrientation3D;
import RTC.TimedFloatSeq;
import RTC.TimedString;
import jp.go.aist.rtm.RTC.DataFlowComponentBase;
import jp.go.aist.rtm.RTC.Manager;
import jp.go.aist.rtm.RTC.port.InPort;
import jp.go.aist.rtm.RTC.port.OutPort;
import jp.go.aist.rtm.RTC.util.DataRef;
import jp.go.aist.rtm.RTC.util.StringHolder;
import RTC.ReturnCode_t;
import gcs.api.VehicleControllerFactoryImpl;
import ms.Uav;
import ms.aif.Connection;
import ms.aif.TelemetryListener;
import ms.aif.VehicleController;
import ms.aif.VehicleControllerFactory;
import ms.aif.command.Control;
import ms.aif.command.ControlCommand;
import ms.aif.command.SingleWayPoint;
import ms.aif.command.WayPointCommand;
import ms.aif.response.CommandResponse;
import ms.aif.response.Status;
import ms.aif.response.TelemetryResponse;

/*!
 * @class MiniSurveyorIOImpl
 * @brief MiniSurveyor flight instruction. Get and output the sensor info.
 *
 * Refer to the "機能仕様書_MiniSurveyorIO_RTC.pdf" for details
 *
 * Refer to the "機能仕様書_MiniSurveyorIO_RTC.pdf" for details
 *
 * Refer to the "機能仕様書_MiniSurveyorIO_RTC.pdf" for details
 *
 */
public class MiniSV06LAImpl extends DataFlowComponentBase {
	// Logger
	private static final java.util.logging.Logger LOGGER = java.util.logging.Logger
			.getLogger(MiniSV06LAImpl.class.getName());

	/*--- Error message ---*/
	private static final String ERR_MSG_1 = "[E]home_location is none. Please \"Exit\" this RTC.";
	private static final String ERR_MSG_2 = "[E]home_location is not exist. Please \"Reset\" this RTC.";
	private static final String ERR_MSG_3 = "[E]No armed. Please \"Reset\" this RTC.";
	private static final String ERR_MSG_4 = "[E]Communication with the GCS has been lost. Start the RTL.";

	/*--- Drone RTC inside status ---*/
	private static final int STS_DISARM = 0;
	private static final int STS_ARMED = 1;
	private static final int STS_TAKEOFF = 2;
	private static final int STS_RTL = 3;
	private static final int STS_IN_MOVE = 4;
	private static final int STS_ALT_CHG = 5;
	private static final int STS_HOVERING = 6;
	private static final int STS_YAW_CHG = 7;
	private static final String[] STS_LIST = { "Disarm", "Armed", "Takeoff",
			"RTL", "IN_Move", "Alt_CHG", "Hovering", "Yaw_CHG" };

	/*--- Flight Mode ---*/
	/* --- getCtrl --- */
	private static final short FM_NO_DATA = 0;
	private static final short FM_MANUAL = 1;
	private static final short FM_ATT = 2;
	private static final short FM_GPSA = 3;
	private static final short FM_NAVI = 4;
	private static final short FM_TAKEOFF = 5;
	private static final short FM_LAND = 6;
	private static final HashMap<Short, String> CTRL_FLIGHT_LIST = new HashMap<Short, String>() {
		{
			put(FM_NO_DATA, "NO_DATA");
			put(FM_MANUAL, "Manual");
			put(FM_ATT, "Assist(Gyro)");
			put(FM_GPSA, "Assist(GPS)");
			put(FM_NAVI, "Navi");
			put(FM_TAKEOFF, "Auto-takeoff");
			put(FM_LAND, "Auto-landing");
		}
	};
	/*--- getAuto ---*/
	private static final short FM_COMMAND = 1;
	private static final short FM_PLAN = 2;
	private static final short FM_GOHOME = 3;
	private static final HashMap<Short, String> AUTO_FLIGHT_LIST = new HashMap<Short, String>() {
		{
			put(FM_NO_DATA, "NAVI");
			put(FM_COMMAND, "Navi(Command)");
			put(FM_PLAN, "Plan");
			put(FM_GOHOME, "Navi(GoHome)");
		}
	};
	/*--- getDrv ---*/
	private static final short FM_DISARM = 0;
	private static final short FM_ARMED = 1;
	private static final short FM_FLIGHT = 2;

	/*--- communication err value 100ms period count times---*/
	private static final short COMM_ERR_TIMEUP = 30;

	/*--- home altitude decrease value [m] ---*/
	private static final Double HOME_DECREASE = 10.0;

	/*--- means minimum move distance [m] ---*/
	private static final Double MIN_MOVE_DIS = 2.0;

	/*--- move speed [m/s] ---*/
	private static final Double MOVE_SPEED = 1.0;

	/*--- home arrival near value ---*/
	private static final Double HOME_ARRIVAL_VAL = 1.0;

	// con
	private Connection con;
	// st
	private Status st;
	// vCon
	private VehicleController vController;
	// uav
	private Uav uav;
	// home
	private GPSData home = new GPSData();
	// drone
	private GPSData drone = new GPSData();
	// vehicle
	private TimedString vehicle = new TimedString();
	// voltage
	private TimedFloat voltage = new TimedFloat();
	// gyro
	private TimedOrientation3D gyro = new TimedOrientation3D();
	// accel
	private TimedFloatSeq accel = new TimedFloatSeq();
	// mag
	private TimedFloat mag = new TimedFloat();
	// time
	private RTC.Time rtcTime = new RTC.Time(0, 0);
	// gpstime
	private RTC.GPSTime gpsTime = new RTC.GPSTime(0, 0);

	/*--- error flag true:safe false:error ---*/
	private Boolean rtc_flg = true;

	/*--- drv vehicle state ---*/
	private Short drv_state = FM_DISARM;
	private Short vehicle_state = FM_NO_DATA;

	/*--- drone state ---*/
	private String drone_state = STS_LIST[STS_DISARM];

	/*--- flight timer ---*/
	private long flight_time_start = 0;

	/*--- err timer ---*/
	private short comm_err_timer = 0;

	/*--- alt change ---*/
	private Double alt_chg_req = 0.0;

	/*--- move target ---*/
	private Double move_target_lat = 0.0;
	private Double move_target_lon = 0.0;

	/*
	 * !
	 * 
	 * @brief constructor
	 * 
	 * @param manager Maneger Object
	 */
	public MiniSV06LAImpl(Manager manager) {
		super(manager);
		// <rtc-template block="initializer">
        m_inTargetPos_val = new GPSData();
        initializeParam(m_inTargetPos_val);
        m_inTargetPos = new DataRef<GPSData>(m_inTargetPos_val);
        m_inTargetPosIn = new InPort<GPSData>("inTargetPos", m_inTargetPos);
        m_outGPSHome_val = new GPSData();
        m_outGPSHome = new DataRef<GPSData>(m_outGPSHome_val);
        m_outGPSHomeOut = new OutPort<GPSData>("outGPSHome", m_outGPSHome);
        m_outGPSDrone_val = new GPSData();
        m_outGPSDrone = new DataRef<GPSData>(m_outGPSDrone_val);
        m_outGPSDroneOut = new OutPort<GPSData>("outGPSDrone", m_outGPSDrone);
        m_outBattV_val = new TimedFloat();
        m_outBattV = new DataRef<TimedFloat>(m_outBattV_val);
        m_outBattVOut = new OutPort<TimedFloat>("outBattV", m_outBattV);
        m_outFlightTime_val = new Time();
        m_outFlightTime = new DataRef<Time>(m_outFlightTime_val);
        m_outFlightTimeOut = new OutPort<Time>("outFlightTime", m_outFlightTime);
        m_outGyro_val = new TimedOrientation3D();
        m_outGyro = new DataRef<TimedOrientation3D>(m_outGyro_val);
        m_outGyroOut = new OutPort<TimedOrientation3D>("outGyro", m_outGyro);
        m_outAccel_val = new TimedFloatSeq();
        m_outAccel = new DataRef<TimedFloatSeq>(m_outAccel_val);
        m_outAccelOut = new OutPort<TimedFloatSeq>("outAccel", m_outAccel);
        m_outMag_val = new TimedFloat();
        m_outMag = new DataRef<TimedFloat>(m_outMag_val);
        m_outMagOut = new OutPort<TimedFloat>("outMag", m_outMag);
        m_outVehicleMode_val = new TimedString();
        m_outVehicleMode = new DataRef<TimedString>(m_outVehicleMode_val);
        m_outVehicleModeOut = new OutPort<TimedString>("outVehicleMode", m_outVehicleMode);
        m_outState_val = new TimedString();
        m_outState = new DataRef<TimedString>(m_outState_val);
        m_outStateOut = new OutPort<TimedString>("outState", m_outState);
        m_outSysErr_val = new TimedString();
        m_outSysErr = new DataRef<TimedString>(m_outSysErr_val);
        m_outSysErrOut = new OutPort<TimedString>("outSysErr", m_outSysErr);
		// </rtc-template>

	}

	/*
	 * !
	 * 
	 * The initialize action (on CREATED->ALIVE transition) formaer
	 * rtc_init_entry()
	 * 
	 * @return RTC::ReturnCode_t
	 */
	@Override
	protected ReturnCode_t onInitialize() {
		// Registration: InPort/OutPort/Service
		// <rtc-template block="registration">
        // Set InPort buffers
        addInPort("inTargetPos", m_inTargetPosIn);
        
        // Set OutPort buffer
        addOutPort("outGPSHome", m_outGPSHomeOut);
        addOutPort("outGPSDrone", m_outGPSDroneOut);
        addOutPort("outBattV", m_outBattVOut);
        addOutPort("outFlightTime", m_outFlightTimeOut);
        addOutPort("outGyro", m_outGyroOut);
        addOutPort("outAccel", m_outAccelOut);
        addOutPort("outMag", m_outMagOut);
        addOutPort("outVehicleMode", m_outVehicleModeOut);
        addOutPort("outState", m_outStateOut);
        addOutPort("outSysErr", m_outSysErrOut);
		// </rtc-template>
		bindParameter("ConnectPort", m_ConnectPort, "COM3");

		this.gyro.data = new Orientation3D();
		this.accel.data = new float[3];
		this.m_outGPSHome_val.altitude = Double.NaN;
		this.m_outGPSHome_val.latitude = Double.NaN;
		this.m_outGPSHome_val.longitude = Double.NaN;
		
		return super.onInitialize();
	}

	/***
	 *
	 * The finalize action (on ALIVE->END transition) formaer
	 * rtc_exiting_entry()
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	protected ReturnCode_t onFinalize() {
		try {
			this.vController.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return super.onFinalize();
	}

	/***
	 *
	 * The startup action when ExecutionContext startup former
	 * rtc_starting_entry()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	protected ReturnCode_t onStartup(int ec_id) {
		getAPIService();
		if (checkConnectStatus(this.con) == false) {
			return ReturnCode_t.RTC_ERROR;
		}

		getTelemetryListener();
		if (checkTelemetryStatus(this.st) == false) {
			return ReturnCode_t.RTC_ERROR;
		}
		
		startSleep(1000);

		if (checkGPSHome() == false) {
			errMsgWrite(ERR_MSG_1);
			return ReturnCode_t.RTC_ERROR;
		}

		return RTC.ReturnCode_t.RTC_OK;

	}

	/***
	 *
	 * The shutdown action when ExecutionContext stop former
	 * rtc_stopping_entry()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	// @Override
	// protected ReturnCode_t onShutdown(int ec_id) {
	// return super.onShutdown(ec_id);
	// }

	/***
	 *
	 * The activated action (Active state entry action) former
	 * rtc_active_entry()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	protected ReturnCode_t onActivated(int ec_id) {
		if (checkConnectStatus(this.con) == false) {
			return super.on_startup(ec_id);
		}
		if (checkGPSHome() == false) {
			errMsgWrite(ERR_MSG_2);
			return ReturnCode_t.RTC_ERROR;
		}
		if (checkArmed() == false) {
			errMsgWrite(ERR_MSG_3);
			return ReturnCode_t.RTC_ERROR;
		}
		flightTimeStart();
		changeDroneSts(STS_ARMED);

		return ReturnCode_t.RTC_OK;
	}

	/***
	 *
	 * The deactivated action (Active state exit action) former
	 * rtc_active_exit()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	protected ReturnCode_t onDeactivated(int ec_id) {
		startRTL();
		return ReturnCode_t.RTC_OK;
	}

	/***
	 *
	 * The execution action that is invoked periodically former rtc_active_do()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	protected ReturnCode_t onExecute(int ec_id) {
		if (checkArmed() == false) {
			changeDroneSts(STS_DISARM);
			this.m_outStateOut.write();
			return super.deactivate(ec_id);
		}

		if (checkDroneSts(STS_RTL)) {
			if (checkHomeArrival() && this.vehicle_state != FM_LAND) {
				startLand();
			}
		}

		drvStateSelect();
		fltCtrlSelect();

		if (readTargetPos()) {
			this.comm_err_timer = 0;
			flightCtrlCore();
		} else {
			this.comm_err_timer += 1;
			if (this.comm_err_timer == COMM_ERR_TIMEUP) {
				this.comm_err_timer = 0;
				errMsgWrite(ERR_MSG_4);
				changeDroneSts(STS_RTL);
				startGOHOME();
			}
		}

		setFlightTime(flightTimeCount());
		portWrite();

		return ReturnCode_t.RTC_OK;
	}

	/***
	 *
	 * The aborting action when main logic error occurred. former
	 * rtc_aborting_entry()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	// @Override
	// public ReturnCode_t onAborting(int ec_id) {
	// return super.onAborting(ec_id);
	// }

	/***
	 *
	 * The error action in ERROR state former rtc_error_do()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	@Override
	public ReturnCode_t onError(int ec_id) {
		startRTL();
		return super.onError(ec_id);
	}

	/***
	 *
	 * The reset action that is invoked resetting This is same but different the
	 * former rtc_init_entry()
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	// @Override
	// protected ReturnCode_t onReset(int ec_id) {
	// return super.onReset(ec_id);
	// }

	/***
	 *
	 * The state update action that is invoked after onExecute() action no
	 * corresponding operation exists in OpenRTm-aist-0.2.0
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	// @Override
	// protected ReturnCode_t onStateUpdate(int ec_id) {
	// return super.onStateUpdate(ec_id);
	// }

	/***
	 *
	 * The action that is invoked when execution context's rate is changed no
	 * corresponding operation exists in OpenRTm-aist-0.2.0
	 *
	 * @param ec_id
	 *            target ExecutionContext Id
	 *
	 * @return RTC::ReturnCode_t
	 * 
	 * 
	 */
	// @Override
	// protected ReturnCode_t onRateChanged(int ec_id) {
	// return super.onRateChanged(ec_id);
	// }
	//
	// Configuration variable declaration
	// <rtc-template block="config_declare">
    /*!
     * To connecting to the target vehicle
     * - Name: ConnectPort ConnectPort
     * - DefaultValue: COM3
     */
    protected StringHolder m_ConnectPort = new StringHolder();
	// </rtc-template>

	// DataInPort declaration
	// <rtc-template block="inport_declare">
    protected GPSData m_inTargetPos_val;
    protected DataRef<GPSData> m_inTargetPos;
    /*!
     * Drone flight target position
     * - Semantics: -Latitude
     *              -Longitude
     *              -Altitude
     * - Unit: deg
     */
    protected InPort<GPSData> m_inTargetPosIn;

    
	// </rtc-template>

	// DataOutPort declaration
	// <rtc-template block="outport_declare">
    protected GPSData m_outGPSHome_val;
    protected DataRef<GPSData> m_outGPSHome;
    /*!
     * GPS Data for "Home location"
     * - Semantics: -Latitude
     *              -Longitude
     *              -Altitude
     * - Unit: deg
     */
    protected OutPort<GPSData> m_outGPSHomeOut;

    protected GPSData m_outGPSDrone_val;
    protected DataRef<GPSData> m_outGPSDrone;
    /*!
     * GPS Data for "Drone location"
     * - Semantics: -Latitude
     *              -Longitude
     *              -Altitude
     * - Unit: deg
     */
    protected OutPort<GPSData> m_outGPSDroneOut;

    protected TimedFloat m_outBattV_val;
    protected DataRef<TimedFloat> m_outBattV;
    /*!
     * Battery voltage
     * - Unit: V
     */
    protected OutPort<TimedFloat> m_outBattVOut;

    protected Time m_outFlightTime_val;
    protected DataRef<Time> m_outFlightTime;
    /*!
     * Drone flight time
     * - Unit: sec
     */
    protected OutPort<Time> m_outFlightTimeOut;

    protected TimedOrientation3D m_outGyro_val;
    protected DataRef<TimedOrientation3D> m_outGyro;
    /*!
     * 3 axis gyro sensor
     * - Semantics: -Roll angle
     *              -Pitch angle
     *              -Yaw angle
     * - Unit: rad
     */
    protected OutPort<TimedOrientation3D> m_outGyroOut;

    protected TimedFloatSeq m_outAccel_val;
    protected DataRef<TimedFloatSeq> m_outAccel;
    /*!
     * Accelerometer
     * - Semantics: -x
     *              -y
     *              -z
     * - Unit: m/sec
     */
    protected OutPort<TimedFloatSeq> m_outAccelOut;

    protected TimedFloat m_outMag_val;
    protected DataRef<TimedFloat> m_outMag;
    /*!
     * Magnetic Compass(Drone's heading)
     * - Semantics: 0-360
     *              (North is 0,East is 90, South is 180, West is
     *              270)
     * - Unit: deg
     */
    protected OutPort<TimedFloat> m_outMagOut;

    protected TimedString m_outVehicleMode_val;
    protected DataRef<TimedString> m_outVehicleMode;
    /*!
     * Current flight mode
     */
    protected OutPort<TimedString> m_outVehicleModeOut;

    protected TimedString m_outState_val;
    protected DataRef<TimedString> m_outState;
    /*!
     * MiniSurveyor I/O RTC internal mode
     * - Semantics: For detail see the
     *              "機能仕様書_MiniSurveyorIO_RTC.docx"
     */
    protected OutPort<TimedString> m_outStateOut;

    protected TimedString m_outSysErr_val;
    protected DataRef<TimedString> m_outSysErr;
    /*!
     * Drone I/O RTC error info
     * - Unit: For detail see the
     *         "機能仕様書_MiniSurveyorIO_RTC.docx"
     */
    protected OutPort<TimedString> m_outSysErrOut;

    
	// </rtc-template>

	// CORBA Port declaration
	// <rtc-template block="corbaport_declare">
    
	// </rtc-template>

	// Service declaration
	// <rtc-template block="service_declare">
    
	// </rtc-template>

	// Consumer declaration
	// <rtc-template block="consumer_declare">
    
	// </rtc-template>

	private void initializeParam(Object target) {
		Class<?> targetClass = target.getClass();
		ClassLoader loader = target.getClass().getClassLoader();
		//
		Field[] fields = targetClass.getFields();
		for (Field field : fields) {
			if (field.getType().isPrimitive())
				continue;

			try {
				if (field.getType().isArray()) {
					Object arrayValue = null;
					Class<?> clazz = null;
					if (field.getType().getComponentType().isPrimitive()) {
						clazz = field.getType().getComponentType();
					} else {
						clazz = loader.loadClass(field.getType()
								.getComponentType().getName());
					}
					arrayValue = Array.newInstance(clazz, 0);
					field.set(target, arrayValue);

				} else {
					Constructor<?>[] constList = field.getType()
							.getConstructors();
					if (constList.length == 0) {
						Method[] methodList = field.getType().getMethods();
						for (Method method : methodList) {
							if (method.getName().equals("from_int") == false)
								continue;
							Object objFld = method.invoke(target,
									new Object[] { new Integer(0) });
							field.set(target, objFld);
							break;
						}

					} else {
						Class<?> classFld = Class.forName(field.getType()
								.getName(), true, loader);
						Object objFld = classFld.newInstance();
						initializeParam(objFld);
						field.set(target, objFld);
					}
				}
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}
		}
	}

	/*------------- start On function -------------*/
	/*--- onStartup ---*/
	// 接続情報取得
	private Connection getConnection(String port) {
		Connection connect = new Connection(port, 115200, 0, 0, 0, true, 3000);
		return connect;
	}

	// APIサービスインスタンス取得
	private void getAPIService() {
		VehicleControllerFactory vfactory = new VehicleControllerFactoryImpl();
		this.con = getConnection(this.m_ConnectPort.value);
		this.vController = vfactory.createVehicleController(this.con);
		LOGGER.info(this.con.getConnectionStatus().getErrorMessage());
	}

	// 通信モジュール接続確認
	private Boolean checkConnectStatus(Connection connect) {
		Boolean ret = false;
		if (connect.getConnectionStatus().getErrorId() == 1
				&& connect.getConnectionStatus().getErrorCategory() == 1) {
			ret = true;
		}
		return ret;
	}

	// イベントリスナ登録
	private void getTelemetryListener() {
		this.st = this.vController
				.setTelemetryListener(new TelemetryListener() {
					Status st_past = null;

					// データ受信イベント
					@Override
					public void onUpdate(TelemetryResponse telemetryResponse) {
						uav = telemetryResponse.getTelemetry().getUav();
						telemetryEvent(uav);

						// 受信イベントのステータスを取得する
						Status status = telemetryResponse.getStatus();
						if (status != st_past) {
							LOGGER.info("======" + status.getErrorMessage()
									+ "======");
						}
						st_past = status;
					}
				});
		LOGGER.info(this.st.getErrorMessage());
	}

	// テレメトリーイベント
	private void telemetryEvent(Uav uav) {
		// GPSHome
		setGPSHome(readGPSHome(uav));
		// GPSDrone
		setGPSDrone(readGPSDrone(uav));
		// VehicleMode
		setVehicleMode(readVehicleMode(uav));
		// Voltage
		setVoltage(readVoltage(uav));
		// Gyro
		setGyro(readGyro(uav));
		// Accel
		setAccel(readAccel(uav));
		// Mag
		setMag(readMag(uav));
		// Drv
		setDrv(readDrv(uav));
	}

	// イベントリスナ接続確認
	private Boolean checkTelemetryStatus(Status status) {
		Boolean ret = false;
		if (status.getErrorId() == 3 && status.getErrorCategory() == 1) {
			ret = true;
		}
		return ret;
	}

	// GPSHomeが取得できたかチェック
	private Boolean checkGPSHome() {
		Boolean ret = true;
		if (Double.isNaN(m_outGPSHome_val.latitude)
				|| Double.isNaN(m_outGPSHome_val.longitude)
				|| Double.isNaN(m_outGPSHome_val.altitude)) {
			ret = false;
		}
		return ret;
	}

	/*--- onActivate ---*/

	// Armed状態の確認
	private Boolean checkArmed() {
		Boolean ret = false;
		if (drv_state != FM_DISARM) {
			ret = true;
		}
		return ret;
	}

	// フライト時間計測開始
	private void flightTimeStart() {
		flight_time_start = System.currentTimeMillis();
	}

	/*--- onDeactivate ---*/
	// GOHOMEコマンドを送る
	private void startGOHOME() {
		Control command = Control.GOHOME;
		CommandResponse cmdRes = vController.sendCommand(new ControlCommand() {
			@Override
			public Control getControl() {
				return command;
			}
		});
		LOGGER.info(cmdRes.getStatus().getErrorMessage());
	}

	// home-drone間距離取得
	private Double getHomeDronePos() {
		DroneMathmetics dist_cal = new DroneMathmetics();
		Double dist = dist_cal.distance_math(m_outGPSHome_val.longitude,
				m_outGPSDrone_val.longitude, m_outGPSHome_val.latitude,
				m_outGPSDrone_val.latitude, 0.0, 0.0);
		return dist;
	}

	// HOMEについたか確認
	private Boolean checkHomeArrival() {
		Boolean ret = false;
		if (getHomeDronePos() <= HOME_ARRIVAL_VAL) {
			ret = true;
		}
		return ret;
	}

	// LANDコマンドを送る
	private void startLand() {
		Control command = Control.LAND;
		CommandResponse cmdRes = vController.sendCommand(new ControlCommand() {
			@Override
			public Control getControl() {
				return command;
			}
		});
		LOGGER.info(cmdRes.getStatus().getErrorMessage());
	}

	// RTLを開始
	private void startRTL() {
		if (!checkDroneSts(STS_DISARM) && !checkDroneSts(STS_ARMED)
				&& !checkDroneSts(STS_RTL)) {
			changeDroneSts(STS_RTL);
			this.m_outStateOut.write();
			startGOHOME();
			while (checkHomeArrival() == false) {
				startSleep(1000);
				this.m_outVehicleModeOut.write();
			}
			startLand();
			changeDisarm();
		// すでにGOHOMEが実行されている場合
		} else if (checkDroneSts(STS_RTL)) {
			this.m_outStateOut.write();
			while (checkHomeArrival() == false) {
				startSleep(1000);
				this.m_outVehicleModeOut.write();
			}
			startLand();
			changeDisarm();
		}
	}

	// Disarm変更
	private void changeDisarm() {
		while (drv_state != FM_DISARM) {
			startSleep(1000);
			this.m_outVehicleModeOut.write();
		}
		changeDroneSts(STS_DISARM);
		this.m_outStateOut.write();
	}

	/*--- onExecute ---*/
	// TargetPos読み込み
	private Boolean readTargetPos() {
		Boolean ret = false;
		if (m_inTargetPosIn.isNew()) {
			m_inTargetPosIn.read();
			ret = true;
		}
		return ret;
	}

	// DRONE State状態の確認
	private Boolean checkDroneSts(int state) {
		Boolean ret = false;
		if (drone_state.equals(STS_LIST[state])) {
			ret = true;
		}
		return ret;
	}

	// フライト制御
	private void flightCtrlCore() {
		if (drv_state == FM_FLIGHT
				&& this.m_outVehicleMode_val.data.equals(AUTO_FLIGHT_LIST
						.get(FM_COMMAND)) && !checkDroneSts(STS_RTL)
				&& !checkDroneSts(STS_TAKEOFF)) {
			if (homeTargetPosSame()) {
				startGOHOME();
				changeDroneSts(STS_RTL);
			} else {
				setTargetPos();
			}
		}
	}

	// ホーム・ターゲット間判定
	private Boolean homeTargetPosSame() {
		Boolean ret = false;
		if (this.m_inTargetPos.v.latitude == this.m_outGPSHome_val.latitude
				&& this.m_inTargetPos.v.longitude == this.m_outGPSHome_val.longitude
				&& this.m_inTargetPos.v.altitude == this.m_outGPSHome_val.altitude) {
			ret = true;
		}

		return ret;
	}

	// Pauseコマンド送信
	private void startPause() {
		Control command = Control.PAUSE;
		CommandResponse cmdRes = vController.sendCommand(new ControlCommand() {
			@Override
			public Control getControl() {
				return command;
			}
		});
		LOGGER.info(cmdRes.getStatus().getErrorMessage());
	}

	// ターゲットポジションセット
	private void setTargetPos() {
		DroneMathmetics dist_cal = new DroneMathmetics();
		Double dist = dist_cal.distance_math(this.m_outGPSDrone_val.longitude,
				this.m_inTargetPos.v.longitude,
				this.m_outGPSDrone_val.latitude, this.m_inTargetPos.v.latitude,
				0.0, 0.0);

		if (dist <= MIN_MOVE_DIS) {
			if (Math.abs(this.m_inTargetPos.v.altitude
					- this.m_outGPSDrone_val.altitude) > 0) {
				changeDroneSts(STS_ALT_CHG);
				LOGGER.fine("alt change:" + this.m_inTargetPos.v.altitude);
				if(alt_chg_req*0.97 <= this.m_outGPSDrone_val.altitude
						&& this.m_outGPSDrone_val.altitude <= alt_chg_req*1.03){
					setSingleWayPoint();
					this.alt_chg_req = this.m_inTargetPos.v.altitude;
				}
			} else {
				startPause();
				changeDroneSts(STS_HOVERING);
				this.alt_chg_req = this.m_inTargetPos.v.altitude;
			}
		} else {
			this.move_target_lat = this.m_inTargetPos.v.latitude;
			this.move_target_lon = this.m_inTargetPos.v.longitude;
			changeDroneSts(STS_IN_MOVE);
			LOGGER.fine("Move to :" + this.m_inTargetPos.v.latitude + ","
					+ this.m_inTargetPos.v.longitude + ","
					+ this.m_inTargetPos.v.altitude);
			setSingleWayPoint();
		}
	}

	// シングルウェイポイント送信
	private void setSingleWayPoint() {
		SingleWayPoint wayPoint = new SingleWayPoint(
				this.m_inTargetPos.v.latitude, this.m_inTargetPos.v.longitude,
				this.m_inTargetPos.v.altitude, 0.0, MOVE_SPEED);

		CommandResponse cmdRes = vController.sendCommand(new WayPointCommand() {
			@Override
			public SingleWayPoint getWayPoint() {
				return wayPoint;
			}
		});
		LOGGER.info(cmdRes.getStatus().getErrorMessage());
	}

	// ターゲット・ドローン間取得（水平）
	private Double getTargetDronePos() {
		DroneMathmetics dist_cal = new DroneMathmetics();
		Double dist = dist_cal.distance_math(this.move_target_lon,
				m_outGPSDrone_val.longitude, this.move_target_lat,
				m_outGPSDrone_val.latitude, 0.0, 0.0);
		return dist;
	}

	// 駆動モード選択
	private void drvStateSelect() {
		if (this.drv_state == FM_DISARM || this.drv_state == FM_ARMED) {
			changeDroneSts(this.drv_state);
		} else {
			if (this.m_outVehicleMode_val.data.equals(CTRL_FLIGHT_LIST
					.get(FM_TAKEOFF))) {
				changeDroneSts(STS_TAKEOFF);
			} else if (this.m_outVehicleMode_val.data.equals(AUTO_FLIGHT_LIST
					.get(FM_GOHOME))) {
				changeDroneSts(STS_RTL);
			}
		}
	}

	// フライトコントロール選択
	private void fltCtrlSelect() {
		if (checkDroneSts(STS_TAKEOFF)) {
			LOGGER.fine("Current Alt:" + this.m_outGPSDrone_val.altitude);
			if (this.vehicle_state != FM_TAKEOFF) {
				LOGGER.info("takeoff complete");
				changeDroneSts(STS_HOVERING);
			}
		} else if (checkDroneSts(STS_IN_MOVE)) {
			LOGGER.fine("Drone <-> Target dist:" + getTargetDronePos());
			if (getTargetDronePos() <= 1) {
				LOGGER.info("Arrived at the target position");
				changeDroneSts(STS_HOVERING);
			}
		}
	}

	// フライト時間カウント
	private int flightTimeCount() {
		long currentTime = System.currentTimeMillis() - this.flight_time_start;

		return (int) currentTime / 1000;
	}

	// 各種Port書き込み
	private void portWrite() {
		this.m_outGPSHomeOut.write();
		this.m_outGPSDroneOut.write();
		this.m_outBattVOut.write();
		this.m_outGyroOut.write();
		this.m_outAccelOut.write();
		this.m_outMagOut.write();
		this.m_outFlightTimeOut.write();
		this.m_outVehicleModeOut.write();
		this.m_outStateOut.write();
	}

	/*------------- end On function -------------*/

	/*------------- start port read & set -------------*/

	// GPSHome取得
	private GPSData readGPSHome(Uav uav) {
		if (uav.getHome() != null) {
			this.home.latitude = MathConv.nullToNan(uav.getHome().getLat());
			this.home.longitude = MathConv.nullToNan(uav.getHome().getLon());
			this.home.altitude = MathConv.nullToNan(homeAltGround(uav
					.getHome().getAlt()));
		}
		return this.home;
	}

	// GPSHome高度を地表値にする.
	private Double homeAltGround(Double alt) {
		if (alt != null) {
			alt = alt - HOME_DECREASE;
		}
		return alt;
	}

	// GPSHomeセット
	private void setGPSHome(GPSData home) {
		if (home != null) {
			this.m_outGPSHome_val.altitude = home.altitude;
			this.m_outGPSHome_val.fixType = RTC.GPSFixType.GPS_FIX_NORMAL;
			this.m_outGPSHome_val.heading = 0.0;
			this.m_outGPSHome_val.horizontalError = 0.0;
			this.m_outGPSHome_val.horizontalSpeed = 0.0;
			this.m_outGPSHome_val.latitude = home.latitude;
			this.m_outGPSHome_val.longitude = home.longitude;
			this.m_outGPSHome_val.numSatellites = 0;
			this.m_outGPSHome_val.timeFromGPS = gpsTime;
			this.m_outGPSHome_val.tm = rtcTime;
			this.m_outGPSHome_val.verticalError = 0.0;
			this.m_outGPSHome_val.verticalSpeed = 0.0;
		}
	}

	// GPSDrone取得
	private GPSData readGPSDrone(Uav uav) {
		if (uav.getGPS() != null) {
			if (uav.getGPS().getLat() != null && uav.getGPS().getLon() != null
					&& uav.getGPS().getAlt() != null) {
				this.drone.latitude = uav.getGPS().getLat();
				this.drone.longitude = uav.getGPS().getLon();
				this.drone.altitude = uav.getGPS().getAlt();
			}
		}

		return this.drone;
	}

	// GPSDroneセット
	private void setGPSDrone(GPSData drone) {
		if (drone != null) {
			this.m_outGPSDrone_val.altitude = drone.altitude;
			this.m_outGPSDrone_val.fixType = RTC.GPSFixType.GPS_FIX_NORMAL;
			this.m_outGPSDrone_val.heading = 0.0;
			this.m_outGPSDrone_val.horizontalError = 0.0;
			this.m_outGPSDrone_val.horizontalSpeed = 0.0;
			this.m_outGPSDrone_val.latitude = drone.latitude;
			this.m_outGPSDrone_val.longitude = drone.longitude;
			this.m_outGPSDrone_val.numSatellites = 0;
			this.m_outGPSDrone_val.timeFromGPS = gpsTime;
			this.m_outGPSDrone_val.tm = rtcTime;
			this.m_outGPSDrone_val.verticalError = 0.0;
			this.m_outGPSDrone_val.verticalSpeed = 0.0;
		}
	}

	// Voltage取得
	private TimedFloat readVoltage(Uav uav) {
		if (uav.getVoltage() != null) {
			this.voltage.data = MathConv.nullTo0(uav.getVoltage()).floatValue();
		}
		return this.voltage;
	}

	// Voltageセット
	private void setVoltage(TimedFloat voltage) {
		if (voltage != null) {
			this.m_outBattV_val.data = voltage.data;
			this.m_outBattV_val.tm = rtcTime;
		}
	}

	// FlightTimeセット
	private void setFlightTime(int time) {
		this.m_outFlightTime_val.sec = time;
		this.m_outFlightTime_val.nsec = 0;
	}

	// Gyro取得
	private TimedOrientation3D readGyro(Uav uav) {
		if (uav.getIMU() != null) {
			if (uav.getIMU().getAtt() != null
					&& uav.getIMU().getAtt().size() == 3) {
				this.gyro.data.r = MathConv.degToRad(MathConv.nullTo0(uav
						.getIMU().getAtt().get(0)));
				this.gyro.data.p = MathConv.degToRad(MathConv.nullTo0(uav
						.getIMU().getAtt().get(1)));
				this.gyro.data.y = MathConv.degToRad(MathConv
						.changeRound(MathConv.nullTo0(uav.getIMU().getAtt()
								.get(2))));
			}
		}
		return this.gyro;
	}

	// Gyroセット
	private void setGyro(TimedOrientation3D gyro) {
		if (gyro != null) {
			this.m_outGyro_val.data = gyro.data;
			this.m_outGyro_val.tm = rtcTime;
		}
	}

	// Accel取得
	private TimedFloatSeq readAccel(Uav uav) {
		if (uav.getIMU() != null) {
			if (uav.getIMU().getAcc() != null
					&& uav.getIMU().getAcc().size() == 3) {
				this.accel.data[0] = MathConv.nullTo0(
						uav.getIMU().getAcc().get(0)).floatValue();
				this.accel.data[1] = MathConv.nullTo0(
						uav.getIMU().getAcc().get(1)).floatValue();
				this.accel.data[2] = MathConv.nullTo0(
						uav.getIMU().getAcc().get(2)).floatValue();
			}
		}
		return this.accel;
	}

	// Accelセット
	private void setAccel(TimedFloatSeq accel) {
		if (accel != null) {
			this.m_outAccel_val.data = accel.data;
			this.m_outAccel_val.tm = rtcTime;
		}
	}

	// Mag取得
	private TimedFloat readMag(Uav uav) {
		if (uav.getIMU() != null) {
			if (uav.getIMU().getAtt() != null
					&& uav.getIMU().getAtt().size() == 3) {
				this.mag.data = MathConv.changeRound(
						MathConv.nullTo0(uav.getIMU().getAtt().get(2)))
						.floatValue();
			}
		}
		return this.mag;
	}

	// Magセット
	private void setMag(TimedFloat mag) {
		if (mag != null) {
			this.m_outMag_val.data = mag.data;
			this.m_outMag_val.tm = rtcTime;
		}
	}

	// VehicleMode取得
	private TimedString readVehicleMode(Uav uav) {
		if (uav.getSys() != null) {
			if (uav.getSys().getMode() != null) {
				if (uav.getSys().getMode().getCtrl() != null) {
					this.vehicle_state = uav.getSys().getMode().getCtrl();
					this.vehicle.data = selectVehicleCtrl(this.vehicle_state);
					if (this.vehicle_state == FM_NAVI) {
						this.vehicle_state = uav.getSys().getMode().getAuto();
						this.vehicle.data = selectVehicleAuto(this.vehicle_state);
					}
				}
			}
		}

		return this.vehicle;
	}

	// VehicleMode選択(Ctrl)
	private String selectVehicleCtrl(Short flag) {
		return CTRL_FLIGHT_LIST.get(flag);
	}

	// VehicleMode選択(Auto)
	private String selectVehicleAuto(Short flag) {
		return AUTO_FLIGHT_LIST.get(flag);
	}

	// VehicleModeセット
	private void setVehicleMode(TimedString vehicle) {
		if (vehicle != null) {
			this.m_outVehicleMode_val.data = vehicle.data;
			this.m_outVehicleMode_val.tm = rtcTime;
		}
	}

	// Stateセット
	private void setState(String state) {
		this.m_outState_val.data = state;
		this.m_outState_val.tm = rtcTime;
	}

	// SysErrセット
	private void setSysErr(String sysErr) {
		this.m_outSysErr_val.data = sysErr;
		this.m_outSysErr_val.tm = rtcTime;
	}

	// Drv取得
	private Short readDrv(Uav uav) {
		if (uav.getSys() != null) {
			return uav.getSys().getMode().getDrv();
		}
		return drv_state;
	}

	// Drvセット
	private void setDrv(Short drv) {
		if (!drv_state.equals(drv)) {
			drv_state = drv;
		}
	}

	/*------------- end data read & set --------------*/

	/*------------- start etc functions -------------*/

	// エラーメッセージ出力
	private void errMsgWrite(String message) {
		setSysErr(message);
		LOGGER.severe(message);
		this.m_outSysErrOut.write();
	}

	// 処理を一時停止する
	private void startSleep(long msec) {
		try {
			Thread.sleep(msec);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	// ドローンステータス変更
	private void changeDroneSts(int state) {
		drone_state = STS_LIST[state];
		setState(drone_state);
	}

	/*------------- end etc functions -------------*/

}
